// menu.h
// Copyright 2011 Play Mechanix Inc. All Rights Reserved
//
// JAL 31-Mar-2011: Creating a standardized (yet flexible) menu system, inititally for use in diag, but usable elsewhere as needed
// Assuming no menu will need more than four buttons (so that we can map our four cabinet switches to them)

#ifndef MENU_H
#define MENU_H

// this is a more or less arbitrary limit of how deep menus can nest
#define MAX_MENU_LEVELS		10

#define  MAX_MENU_CURSORS	4

#define NUM_MENU_HOVER_GUNS	2

// --------------------------------------------------------------------------
// Menu structure
// --------------------------------------------------------------------------
typedef struct menu_t
{
	char name[PM_NAME_SIZE];							// unique identifier
	char menu_type_name[PM_NAME_SIZE];					// Name of the menu type to use (this will set all callbacks based on the type)
	char layout_namespace[PM_NAME_SIZE];				// Namespace to search for layout data
	char icon_name[PM_NAME_SIZE];						// if this item has an icon, specify here
	char value_text_id[PM_NAME_SIZE];					// if this item should display a value, put its text identifier here
	uns32 priority;										// controls placement of this menu within a parent menu
	char bttn1_icon[PM_NAME_SIZE];						// Icon to use for button 1
	char bttn1_label[PM_NAME_SIZE];						// Label text identifier to use for button 1
	char bttn2_icon[PM_NAME_SIZE];						// Icon to use for button 2
	char bttn2_label[PM_NAME_SIZE];						// Label text identifier to use for button 2
	char bttn3_icon[PM_NAME_SIZE];						// Icon to use for button 3
	char bttn3_label[PM_NAME_SIZE];						// Label text identifier to use for button 3
	char bttn4_icon[PM_NAME_SIZE];						// Icon to use for button 4
	char bttn4_label[PM_NAME_SIZE];						// Label text identifier to use for button 4
	List submenu_list;									// List of submenu items (DO NOT populate - it will be populated on install!)
}menu_t;

// When we create a menu, this is everything we need to know about it
typedef struct menu_instance_t
{
	menu_t *menu;			 			// menu we're showing on screen
	int num_focus_items;				// how many things we can switch to
	int active_focus_item;				// the current thing that has focus
	ObjCPMGuiRes *group;				// object of our layout
	int menu_level_number;				// which menu level this menu is (to determine if this menu window is active)
	procHandle_t proc_handle;			// if the menu item starts a proc, here's its handle
} menu_instance_t;

typedef struct menu_coords_t
{
	int x; // x coord of upper-left bound
	int y; // y coord of upper-left bound
	int w; // width
	int h; // height
}menu_coords_t;

typedef struct menu_details_t
{
	menu_coords_t MenuWindowArea;
	int ScrollingListItemHeight;
	int MaxNumVisibleScrollListItems;
	int ScrollbarTopY;
	int ScrollbarTrackHeight;
	int ScrollbarSliderHeight;
	int ShowCursor;
}menu_details_t;

// --------------------------------------------------------------------------
// Menu types
// --------------------------------------------------------------------------

// TODO: Maybe add Button Icon/Text in here, also?
typedef struct menu_type_t
{
	char name[PM_NAME_SIZE];
	int (*MenuCreateFunc)(menu_t *menu, menu_instance_t *inst, menu_details_t menuDetails);	// Function called when menu is first drawn (draw menu in here)
	void (*OnFocusCallback)(menu_t *menu);						// Callback to run when menu takes focus (start procs, play sounds, etc.)
	void (*LeaveFocusCallback)(menu_t *menu);					// Callback to run when menu loses focus (kill procs, stop sounds, etc.)
	void (*UpdateCallback)(menu_t *menu);						// Callback that runs every frame while menu has focus
	int (*Bttn1Callback)(ObjCPMGuiRes *ocpmg);					// Callback for button 1 (this callback will handle all the button's input, state, visibility, etc.)
	int (*Bttn2Callback)(ObjCPMGuiRes *ocpmg);					// Callback for button 2 (this callback will handle all the button's input, state, visibility, etc.)
	int (*Bttn3Callback)(ObjCPMGuiRes *ocpmg);					// Callback for button 3 (this callback will handle all the button's input, state, visibility, etc.)
	int (*Bttn4Callback)(ObjCPMGuiRes *ocpmg);					// Callback for button 4 (this callback will handle all the button's input, state, visibility, etc.)
	void (*SelectAction)();
} menu_type_t;

// Button types
typedef enum _menu_button_type_e
{
	MENU_BUTTON_SCROLL_UP,
	MENU_BUTTON_SCROLL_DOWN,
	MENU_BUTTON_SELECT,
	MENU_BUTTON_CANCEL,
	NUM_BUTTON_COMMANDS,
} menu_button_type_e;

typedef void (*menu_button_cb_t)(void *data);

//typedef int (*menu_item_visible_func_t)(struct _menu_s *menu);
//typedef int (*menu_item_enabled_func_t)(struct _menu_s *menu);

//typedef void (*menu_item_focus_func_t)(void *me);

// --------------------------------------------------------------------------
// Menu button structure
// --------------------------------------------------------------------------
typedef struct menu_button_t
{
	menu_button_type_e button_type;			// Type of button this is					
	char name[PM_NAME_SIZE];				// unique identifier
	int	buttonIsVisible;					// Is this button visible (i.e. can it be activated?)
	core_input_enum mapped_input;			// Input that is tied to this button
	menu_button_cb_t button_callback;		// Callback when this button is activated
} menu_button_t;




typedef enum menuSystemAction_e
{
	MENU_ACTION_NONE = 0,
	MENU_ACTION_PUSH,
	MENU_ACTION_POP,
}menuSystemAction_e;

EXTERN_C menuSystemAction_e gQueuedMenuAction;


// --------------------------------------------------------------------------
// external function declarations
// --------------------------------------------------------------------------
EXTERN_C int MenuStart(char *root_menu_name,
			  char *starting_focus_menu_name,
			  char *layout_namespace,
			  char *bg_layout_name,
			  char *name,
			  char *cam_name,
			  uns32 color_normal,		// 24-bit color (0xRRGGBB)
			  uns32 color_hover,		// 24-bit color (0xRRGGBB)	
			  uns32 color_active,		// 24-bit color (0xRRGGBB)
			  uns32 color_disabled);	// 24-bit color (0xRRGGBB)

EXTERN_C int MenuInit(void);
EXTERN_C void MenuShutDown(void);
EXTERN_C void MenuFinal(void);


EXTERN_C int MenuInstallMenuType(menu_type_t *menu_type);
EXTERN_C int MenuInstallMenu(menu_t *menu, char *parent_menu_name);

// Menu Creation Functions for designing custom menus
EXTERN_C void MenuEraseTempLayout(cpmg_layout_item *layout_item);
EXTERN_C void MenuBeginTitledMenu(menu_t *menu, menu_instance_t *inst);
EXTERN_C void MenuAddButtonBar(menu_t *menu);
EXTERN_C void MenuAddHelpBox(menu_t *menu);
EXTERN_C int MenuImportLayoutIntoMenuWindow(menu_t *menu, char *layout_name, int *x_min, int *y_min);
EXTERN_C int MenuCreateInfoMenuFromLayouts(menu_t *menu, menu_instance_t *inst, menu_details_t menuDetails, char **layouts);

EXTERN_C void MenuBack(void);

// Standard button bar button callbacks
EXTERN_C int MenuCreateScrollingListMenu(menu_t *menu, menu_instance_t *inst, menu_details_t menuDetails);
EXTERN_C int MenuScrollingListMenuUpdateCB(menu_t *menu);
EXTERN_C int MenuScrollingListMenuScrollUpBttnCB(ObjCPMGuiRes *ocpmg);
EXTERN_C int MenuScrollingListMenuScrollDownBttnCB(ObjCPMGuiRes *ocpmg);
EXTERN_C int MenuScrollingListMeunSelectBase(ObjCPMGuiRes *ocpmg);
EXTERN_C int MenuScrollingListMenuSelectBttnCB(ObjCPMGuiRes *ocpmg);
EXTERN_C int MenuScrollingListMenuBackBttnCB(ObjCPMGuiRes *ocpmg);

EXTERN_C menu_instance_t *MenuGetCurMenuInstance(void);
EXTERN_C int MenuCheckObjCPMGCollision(ObjCPMGuiRes *ocpmg, int* pnum);
EXTERN_C void MenuChangeItemColorBasedOnState(ObjCPMGuiRes *item_ocpmg, int state);

EXTERN_C void MenuScrollingListActionSelect();

EXTERN_C void MenuChangeSubitemColorBasedOnState(ObjCPMGuiRes *item_ocpmg, char *subitem, int state);

EXTERN_C void MenuSetCursor(uns32 oid, uns32 pnum);
EXTERN_C void MenuShowAllCursors();
EXTERN_C void MenuHideAllCursors();
EXTERN_C void MenuShowCursor(uns32 gnum);
EXTERN_C void MenuHideCursor(uns32 gnum);

EXTERN_C void MenuDisableAnalogInput(void);
EXTERN_C void MenuEnableAnalogInput(void);

EXTERN_C int PushMenuToTopByName(char *menu_name);


#endif //MENU_H

// EOF

