#ifndef NET_H_
#define NET_H_


#if defined(_WIN32)
	#include "net_win.h"
#else //Linux
	#include "net_linux.h"
#endif //_WIN32


#define NET_MAXSTRINGLEN 128

#define NET_MAX_BUF_SIZE	65536

#define NET_MAX_IP_STR_LEN	80

// errors
#define NETERR_OK			0x80000000	// all OK
#define NETERR_ERROR		0x80000001	// generic error
#define NETERR_SOCKCREATE	0x80000002	// socket() failed
#define NETERR_SOCKCLOSED	0x80000003	// socket was closed
#define NETERR_CKSUMERR		0x80000004	// checksum error
#define NETERR_EOPERR		0x80000005	// EOP not found
#define NETERR_ISCONN		0x80000006	// networking already connected
#define NETERR_NOTCONN		0x80000007	// not connected

// flags
#define NC_SOCKETED			0x0001		// socket exists
#define NC_LISTENING		0x0002		// listening for connections
#define NC_CONNECTED		0x0004		// connected
#define NC_BLOCK			0x0008		// blocking socket (default is non-blocking)
#define NC_BROADCAST		0x0010		// support broadcasting
#define NC_RELIABLEUDP		0x0020		// support reliable udp
#define NC_EXPECTDROP		0x0040		// We expect this net connection to drop

// nc types
#define NC_STREAM			0x0001		// TCP/IP
#define NC_DGRAM			0x0002		// UDP


struct rtudphdr
{
	uns16 flagsize;						// flags.4 size.12
	uns16 sequence;						// sequence #
};

// packet cache
struct rtpacketcache
{
	SocketHandle socket;							// socket this info was sent in
	SocketAddrIn saddr;						// saddr it was sent to

	void* data;								// data that was sent (rtudphdr.32 data.X)

	uns32 time;							// time packet was added
	uns32 retrytime;						// time to retry send
	int nretry;							// # of retries remaining
};

// ack cache
struct rtackcache
{
	uns32 saddr;							// sin_addr.s_addr of host we acked
	uns16 sequence;						// sequence # that was acknowledged
	uns16 unused;
	uns32 expiretime;
};


struct NetConn
{
	NetConn* next;				// l2list pointers
	NetConn* prev;

	char name[128];					// computer name

	uns16 type;						// connection type
	uns16 flags;						// flags
	int id;							// id
	int nframes;						// # of frames since last net data received
	uns32 uptime;						// last update time

	SocketHandle sockfd;						// socket file descriptor
	char ip[16];						// ip address
	uns16 port;						// port
	SocketAddrIn saddr;			// socket address

	//
	uns8* inbuf;						// incoming buffer
	uns16 inbufidx;					// current inbuf index
	uns16 inbuflen;					// length of inbuf
	uns16 inbufsize;					// total size of inbuf

	uns8* outbuf;						// outgoing buffer
	uns16 outbuflen;					// length of outbuf
	uns16 outbufsize;					// total size of outbuf

	uns8* recvbufmem;					// receive buffer allocated memory (we swap pointers during runtime)
	uns8* recvbuf;						// receive buffer
	uns16 recvbuflen;					// current recv buf length
	uns16 recvbufsize;					// total size of recv buf

	//
	int serverdelta;					// time delta (ticks)
	uns32 bytesin;						// # of bytes received
	uns32 bytesout;					// # of bytes sent

	// reliable udp
	uns16 sequence;					// next sequence #
	int npacketcache;				// reliable packet cache
	int packetcachesize;
	rtpacketcache* packetcache;
	int nackcache;					// ack cache
	int ackcachesize;
	rtackcache* ackcache;

	union
	{
		void* data;
		uns32 udata;
		int idata;
	};
};

// net command table
struct NetCommand
{
	char desc[32];					// description
	uns16 cmd;							// command word
	void (*fn)(NetConn* nc);			// receive function
};

// NetCommand table entry
#define NCTBL(desc,cmd,rout) {desc, cmd, rout}

extern int g_netcmdidx;

#ifdef __cplusplus
extern "C" {
#endif
	extern char g_hostname[256];
#ifdef __cplusplus
}
#endif

extern SocketAddrIn g_udpfrom;

// sup
int Net_InitSockets();
void Net_ShutDownSockets();
NetConn* Net_AllocNC(int inbufsize, int outbufsize, uns16 flags);
int Net_FreeNC(NetConn* nc);
int Net_ProcessNC(NetConn* nc, struct NetCommand* cmd, int maxcmd);

// TCP routines
int Net_Connect(NetConn* nc);
int Net_Disconnect(NetConn* nc);

int Net_Listen( NetConn* nc, int portRange );
NetConn* Net_Accept( NetConn* servernc, int inbufsize, int outbufsize );

int Net_Recv(NetConn* nc);
int Net_Send(NetConn* nc);

bool Net_SocketIsValid( SocketHandle sh );
void Net_CloseSocket( SocketHandle sh );
int Net_MakeSocketNonBlocking( SocketHandle sh );

bool Net_WouldBlock();

//
// UDP routines
//

#define UDPF_RELIABLE		0x8000
#define UDPF_ORDERED		0x4000
#define UDPF_UNUSED1		0x2000
#define UDPF_UNUSED2		0x1000
#define UDPF_QOSFLAGS		0xf000

#define UDPH_SIZE(uh)		(uh->flagsize&~UDPF_QOSFLAGS)
#define UDPH_QOS(uh, flag)	(uh->flagsize&flag)

int Net_CreateUDPSocket(NetConn* nc);
int Net_SendToIP(NetConn* nc, char* ip, uns32 flags);
int Net_SendToSADDR(NetConn* nc, SocketAddrIn* saddr, uns32 flags);
int Net_RecvFrom(NetConn* nc);

#define Net_Broadcast(nc) Net_SendToIP(nc, NULL, 0)
int Net_ReliableBroadcastIP(NetConn* nc, char** ips);
int Net_ReliableBroadcastSADDR(NetConn* nc, SocketAddrIn** saddrs);
void Net_CheckUDPPacketCache(NetConn* nc);
void Net_CheckUDPAckCache(NetConn* nc);

// NetConn data routines
uns8 Net_GetByte(NetConn* nc);
void Net_AddByte(NetConn* nc, uns8 b);
uns16 Net_GetWord(NetConn* nc);
void Net_AddWord(NetConn* nc, uns16 a);
uns32 Net_GetDWord(NetConn* nc);
void Net_AddDWord(NetConn* nc, uns32 a);
float Net_GetFloat(NetConn* nc);
void Net_AddFloat(NetConn* nc, float a);
float Net_GetFloat16(NetConn* nc, float min, float max);
void Net_AddFloat16(NetConn* nc, float f, float min, float max);

// get/add data on the network stream.
void Net_GetVec(NetConn* nc, float* v);
void Net_AddVec(NetConn* nc, float* v);
uns32 Net_AddPackedVec3(NetConn* nc, float* v, float xrange, float yrange, float zrange);
void Net_GetPackedVec3(NetConn* nc, float* v, float xrange, float yrange, float zrange);
void Net_GetMat4(NetConn* nc, float* m);
void Net_AddMat4(NetConn* nc, float* m);
void Net_AddString(NetConn* nc, char* str);
void Net_GetString(NetConn* nc, char* str);
void Net_AddMem(NetConn* nc, void* mem, int size);
void Net_GetMem(NetConn* nc, void* mem, int size);

void Net_StartCmd( NetConn* nc, uns8 cmd );
void Net_EndCmd( NetConn* nc );
void Net_SkipCmd( NetConn* nc );

// error checking/log routines
void Net_Error(int32 errid);

//----------------------------------------------------------------------------
// NetMon - Network Monitor infrastructure
//----------------------------------------------------------------------------

// udp commands
enum NMCmd
{
	NMCMD_PING=0,						// YO! YES?
	NMCMD_SCAN,							// am i alive?
	NMCMD_ALIVE,						// yes!
	NMCMD_USER,							// user function
	NMCMD_STRING,						// send string
	NMCMD_SUBSCRIBE,					// subscribe to netmon outpu
	NMCMD_MAX
};

extern NetConn* g_nmudp;

int NetMon_Init(uns16 port);
void NetMon_Shutdown();
int NetMon_Tick();
void NetMon_SetUserFunc(void (*func)(NetConn* nc));
void NetMon_SendString(NetConn* nc, char* format, ...);
void NetMon_SendMsg(char* msg);


#endif