#ifndef PROC_H
#define PROC_H

// proc.h
// Copyright (c) 2010 Play Mechanix Inc. All Rights Reserved
//
// KPJ 29-Jun-2010:  starting almost-complete rewrite of proc

#include <setjmp.h>

// --------------------------------------------------------------------------
// constant definitions
// --------------------------------------------------------------------------

// how long proc names are allowed to be
// it's generally really 1 less than this because it'll force null-termination
#define PROC_REC_MAX_NAME_SIZE	32

// how many active procs we're allowed to have.  this can be pretty much
// anything, it's just here to bound the array
#define PROC_COUNT_MAX			64

// approximately how long it is between frames, and thus about how long each
// frame of sleep is.
#define PROC_FRAME_TIME			PM_FRAME_TIME

// --------------------------------------------------------------------------
// states procs can be in
// --------------------------------------------------------------------------

// type to save our state bits in
typedef uns32 procState_t;

#define PROC_STATE_NULL				0				// nobody's home

#define PROC_STATE_ALIVE			(1 << 0)		// we're a valid proc
#define PROC_STATE_RUNNING			(1 << 1)		// we've been started
#define PROC_STATE_DYING			(1 << 2)		// kill function is running
#define PROC_STATE_DEAD				(1 << 3)		// killed or ended
#define PROC_STATE_CURRENT			(1 << 4)		// running right now
#define PROC_STATE_SLEEP_FRAMES		(1 << 5)		// sleeping for # frames
#define PROC_STATE_SLEEP_GAMETIME	(1 << 6)		// sleeping for gametime
#define PROC_STATE_SLEEP_REALTIME	(1 << 7)		// sleeping in realtime
//#define PROC_STATE_

#define PROC_STATE_SLEEP_MASK		(PROC_STATE_SLEEP_FRAMES | \
									 PROC_STATE_SLEEP_GAMETIME | \
									 PROC_STATE_SLEEP_REALTIME)

#define PROC_STATE_MASK				((1 << 8) - 1)	// all states defined

// --------------------------------------------------------------------------
// flags procs can use
// --------------------------------------------------------------------------

// type to save our flag bits in
typedef uns32 procFlags_t;

#define PROC_FLAG_NULL				0				// we're nothing special

#define PROC_FLAG_ALLOW_DUPES		(1 << 0)		// >1 of this type allowed
#define PROC_FLAG_DUPE_RESTART		(1 << 1)		// create kills old & starts new
#define PROC_FLAG_DUPE_ERROR		(1 << 2)		// breakpoint if this proc gets recreated
#define PROC_FLAG_ARGS_OPTIONAL		(1 << 3)		// no error if args are NULL
#define PROC_FLAG_SURVIVE_WIPEOUT	(1 << 4)		// wipeout doesn't kill this proc
//#define PROC_FLAG_

#define PROC_FLAG_MASK				((1 << 5) - 1)	// all flags defined

// --------------------------------------------------------------------------
// procRec-related typedefs
// --------------------------------------------------------------------------

// datatype for IDs (int because they would typically be based off of an enum)
typedef int procID_t;

// datatype for proc handles
typedef uns32 procHandle_t;

// datatype for proc priority
typedef uns32 procPriority_t;

// function type for start functions (these get called immediately on proc
// creation)
typedef int(*procStartFunc_t)(void *param_data);

// function type for proc functions (what actually runs)
typedef void(*procFunc_t)();

// function type for kill functions (get called when procs get killed/die)
typedef void(*procKillFunc_t)(void *param_data);

// --------------------------------------------------------------------------
// proc start definitions
// --------------------------------------------------------------------------

typedef struct _procStartDef_s
{
//	procID_t id;
	char name[PROC_REC_MAX_NAME_SIZE];
	procStartFunc_t start_func;
	procFunc_t proc_func;
	procKillFunc_t kill_func;
	char params_type[RTDT_MAX_FORM_NAME_LEN];
	procPriority_t priority;
	procFlags_t flags;
} procStartDef_t;

// --------------------------------------------------------------------------
// proc record
// --------------------------------------------------------------------------

// everything we need to know about a proc.
typedef struct _procRec_s
{
//	procID_t number;						// proc ID
	char name[PROC_REC_MAX_NAME_SIZE];		// proc name identifier
	procHandle_t handle;					// handle assigned at create
	procPriority_t priority;				// priority of proc (low = low)

	procState_t state;						// running/dead/sleeping, etc
	procFlags_t flags;						// proc attributes

	time_t start_time;						// unix time proc was created
	uns32 total_elapsed_ms;					// unresettable elapsed millisecs
	uns32 recent_elapsed_ms;				// resettable elapsed millisecs
	uns32 sleep_elapsed_ms;					// elapsed millisecs last sleep
	uns32 total_elapsed_frames;				// unresettable elapsed frames
	uns32 recent_elapsed_frames;			// resettable elapsed frames
	uns32 sleep_elapsed_frames;				// elapsed frames last sleep
	uns32 sleep_amount;						// context-sensitive sleep amount

	procStartFunc_t start_func;				// init func for this proc
	procFunc_t proc_func;					// run func for this proc
	procKillFunc_t kill_func;				// kill func for this proc

	void *stack_begin;						// where stack started for proc
	uns stack_used;							// how many bytes proc has used
	uns stack_save_size;					// how large our save area is
	void *stack_save;						// location where stack is saved

	void *params;							// where our parameters are saved

	jmp_buf env_sys;						// system return point when sleeping
	jmp_buf env_run;						// proc return point when waking
} procRec_t;		// 112 bytes + jmp_bufs = 128 bytes (each is 16 * int)

// --------------------------------------------------------------------------
// externs
// --------------------------------------------------------------------------

// temporary (internal funcs)
extern procRec_t *ProcGetPointerByHandle(procHandle_t ph);
extern procRec_t *ProcGetPointerByName(const char *name);
extern int ProcAssertProcActive(void);
extern int ProcCompareStartDefByName(ListNode *node, void *data);
extern int ProcAddStartDef(procStartDef_t *proc_def);
extern int ProcAddStartDefArray(procStartDef_t *proc_def_array, int count);
extern procHandle_t ProcGetNewHandle(void);
extern procRec_t *ProcGetEmptyProcArraySlot(void);
extern int ProcKillByPointer(procRec_t *proc);

// public interface
// system invocations
extern int ProcInit(void);
extern int ProcReset(uns reset);
extern void ProcFinal(void);
extern int ProcLoop(void);

// within-proc calls
extern procHandle_t ProcGetMyHandle(void);
extern void *ProcGetParams(void);
extern uns32 ProcGetTotalElapsedMs(void);
extern uns32 ProcGetRecentElapsedMs(void);
extern uns32 ProcGetSleepElapsedMs(void);
extern uns32 ProcGetTotalElapsedFrames(void);
extern uns32 ProcGetRecentElapsedFrames(void);
extern uns32 ProcGetSleepElapsedFrames(void);
extern void ProcDoSleepWork(void);
extern void ProcSleepFrames(uns32 frames);
extern void ProcSleepFrametime(uns32 frames);
extern void ProcSleepGametime(uns32 game_ms);
extern void ProcSleepRealtime(uns32 ms);

// this sleeps until game_ms goes by or callback returns non-zero
// returns 0 if slept whole time, else callback return value
extern int ProcSleepGametimeCB( uns32 game_ms, int (*callback)(void *), void *data );


// asking about procs
extern int ProcHandleExists(procHandle_t ph);
extern int ProcNameExists(const char *name);

// manipulation
extern procHandle_t ProcCreate(char *name, void *params);
extern int ProcKill(procHandle_t ph);
extern int ProcKillByName(char *name);
extern int ProcKillAllExceptOneByName(char *name);
extern int ProcKillAllExceptTwoByName(char *name1, char* name2);

extern int ProcWipeout(void);
extern int ProcResetSleeps(void);


// basic input callback
typedef struct
{
	uns32 states;
	int count;
	int *inputs_attract;
} ProcInputCBData;

int ProcInputCB( void *data );

// basic trigger check callback, data should be char *trigger_name
int ProcTriggerCB( void *data );

//extern 

#endif	// PROC_H
