#ifndef _SHADER
#define _SHADER



#include "g5.h"


// SHADER NOTES:



// theory of uniform data blocks
// uniform data blocks map gl buffer objects to collections of uniform variables used in a shader
// this allows uniform vars to be:
// a) set in bulk (save glSetUniform) calls
// b) shared across shaders (same buffer obj can be bound to multiple shaders)
// c) swapped easily (preload several buffers w/ desired data, change binding instead of re-passing data to shader)
// uniform blocks have elements that are both common across all shaders and specific to a given shader
// specifically:
// block data size, member list, member offsets, and member types are all common across all shaders using the same uniform block
// these are used to allocate a buffer object for a block and set the values for it's member data
// block index and member indices are specific to a shader
// these are used to bind a buffer object to uniform block for each specific shader
// our system maintains two sets of data to make all this magic happen:
// with each shader is stored a hash table that maps names to descriptions of the uniform blocks used by that shader
// the decriptions are of type ShaderUniformBlock and contain the shader specific block index and member indices, plus current bind point for the block
// (for completeness they also include copies of common block info (members, offsets, etc)
// Additionally, for each new uniform block name a global ShaderUniformBlockBufferCollection is setup
// the collection contains a header describing the uniform block (size, members, member offsets, types, etc.)
// and a hash table which maps arbitray strings keys to gl buffer objects which are ready to be bound to the block uniform for a shader
// buffer objects are added/removed from the collection with:  ShaderSetUniformBlockBuffer()/ShaderDelUniformBlockBuffer()
// buffer objects are bound with: RenderState_BindUniformBlockBuffer(), ShaderBindUniformBlock()
// additionally ShaderBindUniformBlockToBuffer() wraps both these functions together
// some details about binding:
// uniform blocks are not directly bound to buffer objects
// instead, buffer objects are bound to 'buffer bindings', shader uniform blocks are bound (by block_index) to 'buffer bindings'
// there are fixed number of available binding points
// render state (see render_state.h) tracks which uniform block buffers are bound to each binding point, 
// and the ShaderUniformBlock structs for each shader store what bind point each uniform block on the shaders is bound to
// this level of abstraction has it's uses
// specifically:
// we reserve a set of binding points for some uniform blocks that are directly supported by the engine
// (ex. uFogData and uShadMapData, uLightData, etc.)
// these blocks are bound to pre-defined bind points, the bind points are bound buffers represented the current render state
// in this way, the values for these shared uniforms can be quickly changed across all shaders
// w/o the need to update the binding for each shader individually
// in this vain, certain uniform data block buffers are setup automatically by RenderStateInit()
// and automatically bound and switched by the various render state commands
// all other work with uniform blocks should be done by a shader's setup and/or prerender funcs
//
// nota buena:
// technically GL allows arbitrary mix-matching of uniform block contents/definitions and names across shader programs
// so a block with the same name could have different members in two different shaders
// and two shaders could have blocks with different names but the same contents and thus be bound to the same buffer object
// However for the sake of simplicity and sanity, we force all block names to be global in scope,
// so the definition for uniform block with a given name must match across all shaders
// since block names are arbitrary this does not impose any real restriction on what can be done, just requires that one pay attention
// to the block names used
// 


enum CoreShadersEnum
{
	CORESHADER_SHADERS,
	CORESHADER_SHADERSA,
	CORESHADER_SHADERS_SK,

	CORESHADER_SHADERC,
	CORESHADER_SHADERC_SK,

	CORESHADER_SHADER0,
	CORESHADER_SHADER0_SK,

	CORESHADER_SHADER1,
	CORESHADER_SHADER1_SK,

	CORESHADER_SHADER2,
	CORESHADER_SHADER2_SK,

	CORESHADER_SHADER3,
	CORESHADER_SHADER3_SK,

	CORESHADER_SHADER4,
	CORESHADER_SHADER4_ENVMAP,
	CORESHADER_SHADER4_ENVMAP_SPC,
	CORESHADER_SHADER4_SK,
	CORESHADER_SHADER4_SK_ENVMAP,
	CORESHADER_SHADER4_SK_ENVMAP_SPC,

	CORESHADER_SHADER5,
	CORESHADER_SHADER5_SK,
	CORESHADER_SHADER5P,
	CORESHADER_SHADER5_ENVMAP,
	CORESHADER_SHADER5_ENVMAP_SPC,
	CORESHADER_SHADER5_SK_ENVMAP,
	CORESHADER_SHADER5_SK_ENVMAP_SPC,


	CORESHADER_SHADER6,
	CORESHADER_SHADER6_ENVMAP,
	CORESHADER_SHADER6_ENVMAP_SPC,
	CORESHADER_SHADER6_ENVMAP_ORTHO,
	CORESHADER_SHADER6_ENVMAP_ORTHO_SPC,
//	CORESHADER_SHADER6_SK,

	CORESHADER_SHADER7,
	CORESHADER_SHADER7_SK,

	CORESHADER_SHADER7_PRELIGHT_A,
	CORESHADER_SHADER7_PRELIGHT_B,
	CORESHADER_SHADER7_SK_PRELIGHT_A,
	CORESHADER_SHADER7_SK_PRELIGHT_B,


	CORESHADER_SHADER8,


	// one dir light
	//CORESHADER_SHADER13,
	//CORESHADER_SHADER13_SK,

	/* unused parallax map shaders
	CORESHADER_SHADER9,
	CORESHADER_SHADER9_SK,

	CORESHADER_SHADER11,
	CORESHADER_SHADER11_SK,
	*/

/*
	CORESHADER_SHADER12,
	CORESHADER_SHADER12_SK,

	CORESHADER_SHADER13,
	CORESHADER_SHADER13_SK,

	CORESHADER_SHADER13A,
	CORESHADER_SHADER13A_SK,

	CORESHADER_SHADER13B,
	CORESHADER_SHADER13B_SK,
*/
	CORESHADER_SHADER18,
	CORESHADER_SHADER18_ENVMAP,

	CORESHADER_SHADER21,
	CORESHADER_SHADER21_SK,
	
	CORESHADER_SHADER28,
	CORESHADER_SHADER28_ENVMAP,

	// 3 dir lights
	CORESHADER_SHADER33,
	CORESHADER_SHADER33_SK,

	CORESHADER_SHADER35,
	CORESHADER_SHADER35_SK,
	CORESHADER_SHADER35_ENVMAP,
	CORESHADER_SHADER35_ENVMAP_SPC,
	CORESHADER_SHADER35_SK_ENVMAP,
	CORESHADER_SHADER35_SK_ENVMAP_SPC,

	CORESHADER_SHADER36,
	CORESHADER_SHADER36_ENVMAP,
	CORESHADER_SHADER36_ENVMAP_SPC,

	CORESHADER_SHADER37,
	CORESHADER_SHADER37_SK,

	CORESHADER_SHADER37_PRELIGHT_A,
	CORESHADER_SHADER37_SK_PRELIGHT_A,


	// special fx shaders
	CORESHADER_SHADER100,
	CORESHADER_SHADER100b,
	CORESHADER_SHADER101,
	CORESHADER_SHADER102,
	CORESHADER_SHADER103,
	CORESHADER_SHADER103B,

	CORESHADER_SHADER104,
	CORESHADER_SHADER105,
	CORESHADER_SHADER106,
	CORESHADER_SHADER106_SK,
	CORESHADER_SHADER107,

	CORESHADER_SHADER108,
	CORESHADER_SHADER108b,
//	CORESHADER_SHADER108c,
	CORESHADER_SHADER108c_SK,
	CORESHADER_SHADER108d,
	CORESHADER_SHADER108d_SK,
	CORESHADER_SHADER109,

	CORESHADER_SHADER110,
	CORESHADER_SHADER111,
	CORESHADER_SHADER112,
	CORESHADER_SHADER112_ENVMAP,
	CORESHADER_SHADER112_ENVMAP_SPC,
	CORESHADER_SHADER113,
	CORESHADER_SHADER114,
	CORESHADER_SHADER115,
	CORESHADER_SHADER116,
	
	CORESHADER_SHADER117,
	CORESHADER_SHADER118,
	CORESHADER_SHADER118_SK,
	
	CORESHADER_SHADER119,

	CORESHADER_SHADER_CAR,
	CORESHADER_SHADER_CAR_ENVMAP,
	CORESHADER_SHADER_CAR_ENVMAP_SPC,

	CORESHADER_SHADER_GLASS,
	CORESHADER_SHADER_GLASS_ENVMAP,
	CORESHADER_SHADER_GLASS_ENVMAP_SPC,

	CORESHADER_SHADER_ALPHA_DBG,
	CORESHADER_DEPTHMASK,

	CORESHADER_SHADER_ENVMAP_DYN,
	CORESHADER_SHADER_ENVMAP_DYN_SK,

	CORESHADER_SHADEROVR,

	// game specific shaders
	CORESHADER_SHADER200,  // t4 screen hit fx
	CORESHADER_SHADER201,  // blocks shader
	CORESHADER_SHADER202,  // blocks shader, no tex


	// deffered shaders
	CORESHADER_SHADER305,  
	CORESHADER_SHADER305_SK, 
	CORESHADER_SHADER306,
	CORESHADER_SHADER306_SK,


	NUM_CORESHADERS,

};


// some features are common enough that there are flags
// that instruct engine to supply whatever data they need
// for other features, use shader setup, prerender, render, etc. funcs
#define SHADER_FLAGS_USE_SHADMAPS  0x00000001
#define SHADER_FLAGS_USE_BLOOM     0x00000002
#define SHADER_FLAGS_USE_LIGHTS    0x00000004
#define SHADER_FLAGS_USE_FOG       0x00000008
#define SHADER_FLAGS_USE_SKEL_MATS 0x00000010
#define SHADER_FLAGS_USE_SKEL_MATS_PREV 0x00000020
#define SHADER_FLAGS_USE_GLOW     0x00000040

void *MemAllocShaderLoader( size_t size );

typedef struct 
{
	char name[PM_NAME_SIZE];

	uns32 shaderflags;
	uns8  render_priority; // how to prioritize this shader when sorting objs by shader for rendering (lower = render obj's with this shader first)

	void  (*setup_func)(Res *shader, uns32 renderflags);    // run whenever shader is 'turned on' (set gl states here), input is shader res, shader obj, state rec, flags
	void  (*teardown_func)(Res *shader, uns32 renderflags); // run whenever shader is 'turned off', input is shader res, flags
	void  (*prerender_func)(Res *shader, ObjDispRes *odr, uns32 renderflags); // run just before each obj is rendered (use to set obj specific shader uniform vars), input is shader res, obj disp res, state rec, flags
	void  (*render_func)(Res *shader, ObjDispRes *odr, uns32 renderflags);  // used to render an obj disp res w/ shader (glBegin, glVert, glEnd, etc, goes here), input is shader res, obj disp res, state rec, flags
	//void  (*render_func)( void *shader_res, void *odr, void *sr, uns flags );
	void  (*postrender_func)(Res *shader, ObjDispRes *odr, uns32 renderflags); // run after each odr is rendered, input is shader res, obj disp res, state rec, flags

	char *lightprofile; // ptr to a list (of VID_MAXLIGHTS size) telling what light type of light (RESDATALIGHT_TYPES) the shader supports for each light slot ( -1 for any type of light)

	int load_at_init;        // load shader at init time
	Res *res;  // the shader res
	
	uns32 texUnitsUsed[RENDERSTATE_MAX_TEXS];

} ShaderInfoStruct;

extern ShaderInfoStruct gCoreShaders[NUM_CORESHADERS];



// uniform data stuff
// this is defined here to avoid include conflicts
enum {
	SHADER_DATA_TYPE_UNKNOWN,
	SHADER_DATA_TYPE_INT,
	SHADER_DATA_TYPE_FLOAT,
	SHADER_DATA_TYPE_VEC2,
	SHADER_DATA_TYPE_VEC3,
	SHADER_DATA_TYPE_VEC4,
	SHADER_DATA_TYPE_MAT3X3,
	SHADER_DATA_TYPE_MAT3X4,
	SHADER_DATA_TYPE_MAT4X3,
	SHADER_DATA_TYPE_MAT4X4,
	SHADER_UNIFORMDATA_NUM_TYPES,

};



typedef struct ShaderUniformData {
	char name[PM_NAME_SIZE];
	int location; 
	int type;
	int max_count;
	int count;
	void *values;
} ShaderUniformData;

// for simplicity this next struct is used by both ShaderUniformBlock and ShaderUniformBlockBufferCollection
// just note that the index values are meaningless in the ShaderUniformBlockBufferCollection context
typedef struct ShaderUniformBlockItem {
	char name[PM_NAME_SIZE];
	uns32 index;  // shoudl this be an int?
	uns32 offset;
	uns32 stride;
	int type;
	int count;
} ShaderUniformBlockItem;

typedef struct ShaderUniformBlock {
	char name[PM_NAME_SIZE];
	uns32 block_index; 
	int size;
	int bind_point;
	int item_count;

	ShaderUniformBlockItem *items;

} ShaderUniformBlock;


typedef struct ShaderUniformBlockBufferCollection
{
	int size;
	int item_count;
	ShaderUniformBlockItem *items;
	HashTable item_hash; // items hashed by name (ptrs point items in items list)

	HashTable buffers; // key is string, val is gl buffer object
	HashTable buffers_data; // key is string, val is current contents of gl buffer object
} ShaderUniformBlockBufferCollection;


// model view and texture matrices are special enough that we track
// which variants a shader wants and pass them down automatically
// matrices provided as standalone uniforms
#define SHADER_MATRIX_MODELVIEWPROJ    0x00000001
#define SHADER_MATRIX_MODELVIEW        0x00000002
#define SHADER_MATRIX_MODELVIEW_NORMAL 0x00000004
#define SHADER_MATRIX_PROJECTION       0x00000008

#define SHADER_MATRIX_SHADSPLIT0        0x01000000
#define SHADER_MATRIX_SHADSPLIT1        0x02000000
#define SHADER_MATRIX_SHADSPLIT2        0x04000000

#define SHADER_MATRIX_TEX0             0x00100000
#define SHADER_MATRIX_TEX1             0x00200000
#define SHADER_MATRIX_TEX2             0x00400000
#define SHADER_MATRIX_TEX3             0x00800000

// matrices provided as block data
#define SHADER_MATRIX_STD_MVP_BLOCK    0x10000000
#define SHADER_MATRIX_STD_TEXMAT_BLOCK 0x20000000

// rather than the old 'ClientState' crud
// shaders now just take some set of
// arbitrary data, of course, we standardize a few
#define VERT_ATTRIB_MASK_ARRAY0	0x00000001
#define VERT_ATTRIB_MASK_ARRAY1	0x00000002
#define VERT_ATTRIB_MASK_ARRAY2	0x00000004
#define VERT_ATTRIB_MASK_ARRAY3	0x00000008
#define VERT_ATTRIB_MASK_ARRAY4	0x00000010
#define VERT_ATTRIB_MASK_ARRAY5	0x00000020
#define VERT_ATTRIB_MASK_ARRAY6	0x00000040
#define VERT_ATTRIB_MASK_ARRAY7	0x00000080
#define VERT_ATTRIB_MASK_ARRAY_LAST	VERT_ATTRIB_MASK_ARRAY7
// really it is up to the shader to sort out what attribs it
// wants on what indices, but for sanity's sake we standardize some
#define SHADER_VERT_ATTRIB_INDEX_POS             0
#define SHADER_VERT_ATTRIB_INDEX_COLOR           1
#define SHADER_VERT_ATTRIB_INDEX_UV_SET0         2
#define SHADER_VERT_ATTRIB_INDEX_UV_SET1         3
#define SHADER_VERT_ATTRIB_INDEX_NORMAL          4
#define SHADER_VERT_ATTRIB_INDEX_TANGENT         5
#define SHADER_VERT_ATTRIB_INDEX_SKEL_INDICES    6
#define SHADER_VERT_ATTRIB_INDEX_SKEL_WEIGHTS    7 

#define VERT_ATTRIB_MASK_POS             VERT_ATTRIB_MASK_ARRAY0
#define VERT_ATTRIB_MASK_COLOR           VERT_ATTRIB_MASK_ARRAY1
#define VERT_ATTRIB_MASK_UV_SET0         VERT_ATTRIB_MASK_ARRAY2
#define VERT_ATTRIB_MASK_UV_SET1         VERT_ATTRIB_MASK_ARRAY3
#define VERT_ATTRIB_MASK_NORMAL          VERT_ATTRIB_MASK_ARRAY4
#define VERT_ATTRIB_MASK_TANGENT         VERT_ATTRIB_MASK_ARRAY5
#define VERT_ATTRIB_MASK_SKEL_INDICES    VERT_ATTRIB_MASK_ARRAY6
#define VERT_ATTRIB_MASK_SKEL_WEIGHTS    VERT_ATTRIB_MASK_ARRAY7 


// we define a global struct to contain all our various 'special' uniform data/block names
// this is done just to keep the strings from being splotted out all over the pace
// we also do this a struct instead of a set of const char vars
// to make them easier to extern, and avoid const char horror show
typedef struct
{
	char uFogData[16];
	char uFogColor[16];
	char uFogVals[16];

	char uLightData[12];
	char uLights[12];
	char uAmbient[12];
	char uActiveLights[16];
	char uSpecExponent[16];

	char uBloomThreshold[20];

	char uModelViewProjectionMatrix[32];
	char uModelViewMatrix[24];
	char uModelViewNormalMatrix[24];
	char uStandardModelViewProjMatrixData[36];
	char uProjectionMatrix[24];

	char uStandardTextureMatrixData[36];
	char uTextureMatrix0[16];
	char uTextureMatrix1[16];
	char uTextureMatrix2[16];
	char uTextureMatrix3[16];

	// shad maps
	char uShadMapData[16];
	char uShadParams[12];
	char uShadDir[12];
	char uShadProjMatrix_Split0[32];
	char uShadProjMatrix_Split1[32];
	char uShadProjMatrix_Split2[32];

	char uSkelMatsData[24];
	char uSkelMatsDataPrev[24];
	char uSkelMats[16];
	char uSkelMatsPrev[16];



} StandardUniformNames;

extern StandardUniformNames gStdUniformNames;

// helper structs for casting out the different types of uniform data
typedef struct
{
	int val;
} SUData_int;

typedef struct
{
	float val;
} SUData_float;

typedef struct
{
	float val[2];
} SUData_vec2;


typedef struct
{
	float val[3];
} SUData_vec3;

typedef struct
{
	float val[4];
} SUData_vec4;

typedef struct
{
	float val[9]; // 12];  // gl auto promotes each col to a vec 4
} SUData_mat3x3;

typedef struct
{
	float val[12];
} SUData_mat3x4;


typedef struct
{
	float val[12];
} SUData_mat4x3;


typedef struct
{
	float val[16];
} SUData_mat4x4;

extern int gSUData_Sizes[SHADER_UNIFORMDATA_NUM_TYPES];



ShaderInfoStruct *FindShaderInfoStruct(char *shader_name);
int ShaderSetLoadAtInit( char *shader_name, int load_at_init ); // if passed '*' will apply setting to all core shaders

int UnLoadShaderProgramRes( char *filename ); 
int LoadShaderProgramRes( Res **resp, char *filename, int force_reload ); 
int CreateShaderProgramRes( Res **resp, char *name, char *descr, uns32 shaderprogram );


int ResLoadCoreShaders();



int SetUniformData( ShaderUniformData *uniformdata );
int ShaderSetUniformData( char *name, int type, int count, void *values );
//#define SetUniformDataFloat( name, value )  ShaderSetUniformData( name, SHADER_DATA_TYPE_FLOAT, 1, &value )
int SetUniformDataFloat( char *name, float value );
int SetUniformDataInt( char *name, int value );

int ShaderCheckUniformData( char *name ); // checks if shader has passed uniform data var

// ShaderSetUniformBlockBuffer() notes
// buffer object for buffer_name will be created if not alread exists
// if data == NULL || size <= 0, then buffer is sent w/o changing it (useful for queueing changes to several items before comitting)
int ShaderSetUniformBlockBuffer(  char *block_name, char *buffer_name, void *data, int size );

// ShaderSetUniformBlockBufferItem() notes:
// changes w/o commit_to_udbo are made to local buffer but not comitted to buffer object in opengl space
// useful if making several changes in a row, followed by a ShaderSetUniformBlockBuffer() w/ data==NULL (commits entire buffer)
// although is the changed items are a small part of whole buffer, then maybe faster to just comitt each small change, rather than queue and comitt entire buffer
int ShaderSetUniformBlockBufferItem(  char *block_name, char *buffer_name,  char *item, int type, int count, void *data, int commit_to_udbo );

int ShaderDelUniformBlockBuffer(  char *block_name, char *buffer_name );
ShaderUniformBlockBufferCollection *ShaderGetUniformBlockBufferCollection( char *block_name );

int ShaderBindUniformBlock( char *shader, char *block_name, int bind_point );
int ShaderBindUniformBlockToBuffer( char *shader, char *block_name, char *buffer_name, int bind_point ); // bind_point of -1 means go find one and use that

// shader specific helper funcs

int Shader103GenOffsetsAndCount( int cols, int rows, int texw, int texh );
int Shader103GenGaussianKernel( int cols, int rows, float stdev, float scale );
int Shader103GenLanczosKernel( int cols, int rows, float window, float scale );


int Shader105SetValues( float brightness, float contrast, float *rgb );

int Shader106SetValues( float extrusion, float *rgba );

int ShaderSetActiveLightsFunc(Res *r, uns32 renderflags );



#endif


//EOF

